"""
Contains functionality to transform TinyPseuCo into z3 formulas
"""

from functools import singledispatch, update_wrapper
import itertools
import typing
from typing import List, Tuple

import z3
from z3 import ForAll, And, Or, Implies, Exists

from pseuco_parser import (BaseExpression, UnaryExpression, BinaryExpression,
                           UnaryOperator, BinaryOperator, IntegerConstant,
                           BooleanConstant, VariableUsage, Statement,
                           EmptyStatement, SimpleStatement, MultiStatement,
                           ITE, VariableAssignment, VariableDeclaration,
                           PrintlnStatement, PseuCoType)

__all__ = ["Z3Compiler"]

# Datatypes and predicates needed to encode PseuCo in z3

Instruction = z3.DeclareSort('Instruction')
# we ensure that natural program order is a subset of instruction order by
# construction
InstructionOrder = z3.Function("InstructionOrder", Instruction,
                               Instruction, z3.BoolSort())
PseuCoVar = z3.DeclareSort('PseuCoVar')
# Todo: support more than int for read and write
ReadValue = z3.Function("read", Instruction, PseuCoVar, z3.IntSort(),
                        z3.BoolSort())
WriteValue = z3.Function("write", Instruction, PseuCoVar, z3.IntSort(),
                         z3.BoolSort())
# we need to avoid that ReadValue/WriteValue create read/writes from spurious
# statements which don't actually touch variables

# Many properties only have to hold for statements which are actually reachable
# thus we need to define a predicate for reachability
Reachable = z3.Function("reachable", Instruction, z3.BoolSort())


def methoddispatch(func):
    """ Taken from
    https://stackoverflow.com/questions/24601722/how-can-i-use-functools-singledispatch-with-instance-methods
    Allows single dispatch for methods
    """
    dispatcher = singledispatch(func)

    def wrapper(*args, **kw):
        return dispatcher.dispatch(args[1].__class__)(*args, **kw)
    wrapper.register = dispatcher.register
    update_wrapper(wrapper, func)
    return wrapper


class Z3Compiler:

    def __init__(self):
        self.local_var_table: typing.Mapping[str, z3.ArithRef] = {}
        self.global_var_table: typing.Mapping[str, PseuCoVar] = {}
        self.solver: z3.Solver = z3.Solver()
        self.statement_enablers: List[z3.z3.ExprRef] = []
        self.global_reads: List[Tuple[Instruction, PseuCoVar, z3.IntNumRef]] \
            = []
        self.global_writes: List[Tuple[Instruction, PseuCoVar, z3.IntNumRef]] \
            = []
        self.parent_instructions: List[Instruction] = []
        self.current_agent_id = 0
        self.instructions = []
        self.direct_predecessors = {}

    def make_global_read(self, varname: str,
                         instr: Instruction) -> z3.z3.ArithRef:
        """Takes the variable name and where it is written
        :returns: a variable reperesenting the result of the read
        """
        read_value = z3.Const(str(self.current_agent_id) + str(instr),
                              z3.IntSort())
        variable = self.global_var_table[varname]
        self.global_reads.append((instr, variable, read_value))
        return read_value

    def make_global_write(self,
                          varname: str,
                          instr: Instruction,
                          written_value: z3.z3.ArithRef):
        variable = self.global_var_table[varname]
        self.global_writes.append((instr, variable, written_value))

    def create_global_pseuco_variable(self, var_name):
        var = z3.Const(var_name, PseuCoVar)
        self.global_var_table[var_name] = var
        return var

    def create_local_pseuco_variable(self, name, var_type: PseuCoType):
        if var_type is PseuCoType.Int:
            new_var = z3.Int(str((self.current_agent_id, name)))
        elif var_type is PseuCoType.Boolean:
            new_var = z3.Bool(str((self.current_agent_id, name)))
        else:
            raise NotImplementedError("No support for %s" % str(var_type))
        assert name not in self.local_var_table, "no shadowing!"
        self.local_var_table[name] = new_var
        return new_var

    @methoddispatch
    def compile_expression(self, expression: BaseExpression) -> z3.z3.ExprRef:
        raise NotImplementedError("This does not work")

    @compile_expression.register(UnaryExpression)
    def _(self, unary_exp: UnaryExpression):
        if unary_exp.operator is UnaryOperator.Plus:
            return self.compile_expression(unary_exp.operand)
        elif unary_exp.operator is UnaryOperator.Minus:
            return -self.compile_expression(unary_exp.operand)
        elif unary_exp.operator is UnaryOperator.Not:
            return not self.compile_expression(unary_exp.operand)
        else:
            raise NotImplementedError("%s is not supported" %
                                      unary_exp.operator)

    @compile_expression.register(IntegerConstant)
    def _(self, c: IntegerConstant) -> z3.z3.IntNumRef:
        return z3.IntVal(c.value)

    @compile_expression.register(BooleanConstant)
    def _(self, c: BooleanConstant) -> z3.z3.BoolRef:
        return z3.BoolVal(c.value)

    @compile_expression.register(BinaryExpression)
    def _(self, binary_exp: BinaryExpression):
        lhs = self.compile_expression(binary_exp.lhs)
        rhs = self.compile_expression(binary_exp.rhs)
        op = binary_exp.operator
        if op is BinaryOperator.Plus:
            return lhs + rhs
        elif op is BinaryOperator.Minus:
            return lhs - rhs
        elif op is BinaryOperator.Mult:
            return lhs * rhs
        elif op is BinaryOperator.Div:
            return lhs / rhs
        elif op is BinaryOperator.And:
            return lhs and rhs
        elif op is BinaryOperator.Or:
            return lhs or rhs
        elif op is BinaryOperator.Equal:
            return lhs == rhs
        elif op is BinaryOperator.NEqual:
            return lhs != rhs
        elif op is BinaryOperator.Lt:
            return lhs < rhs
        elif op is BinaryOperator.Lte:
            return lhs <= rhs
        elif op is BinaryOperator.Gt:
            return lhs > rhs
        elif op is BinaryOperator.Gte:
            return lhs >= rhs

    @compile_expression.register(VariableUsage)
    def _(self, var: VariableUsage):
        # TODO: how to handle the global?
        # Plan: make_global_read returns the Int where the current value of the
        # read is stored
        # Before calling compile_expresion, we update the local_var_table so
        # that varname points to the freshly read value
        return self.local_var_table[var.varname]

    def _create_and_order_instruction(self,
                                      statement: Statement) -> Instruction:
        instr = z3.Const(str((self.current_agent_id,
                              statement.id,
                              str(statement)
                              )),
                         Instruction)
        # make sure that every instruction is dinstinct from any other
        # instruction
        for existing_instr in self.instructions:
            self.solver.add(instr != existing_instr)
        self.instructions.append(instr)
        for parent in self.parent_instructions:
            # self.solver.add(InstructionOrder(parent, instr))
            self.solver.assert_and_track(InstructionOrder(parent, instr),
                                         "%s < %s" % (str(parent), str(instr)))
        self.parent_instructions = [instr]
        self.solver.assert_and_track(And(*self.statement_enablers) ==
                                     Reachable(instr),
                                     "%s <==> reachable %s" %
                                     (str(self.statement_enablers),
                                      str(instr)
                                      )
                                     )
        return instr

    def _handle_globals_in_read(self,
                                instr: Instruction,
                                expression: BaseExpression):
        for varname in expression.read_global_variables:
            val = self.make_global_read(varname, instr)
            self.local_var_table[varname] = val

    @methoddispatch
    def compile_statement(self, statement: Statement):
        raise NotImplementedError("no support for %s" % str(statement))

    @compile_statement.register(EmptyStatement)
    def _(self, statement: EmptyStatement):
        return

    @compile_statement.register(SimpleStatement)
    def _(self, statement: SimpleStatement):
        if statement.expression.read_global_variables:
            # if the statement affects global variables, we need to order the
            # corresponding instruction
            instr = self._create_and_order_instruction(statement)
            self._handle_globals_in_read(instr, statement.expression)

    @compile_statement.register(PrintlnStatement)
    def _(self, println: PrintlnStatement):
        # TODO: handle println which prints more
        if println.expressions[0].read_global_variables:
            instr = self._create_and_order_instruction(println)
            self._handle_globals_in_read(instr, println.expression)

    @compile_statement.register(MultiStatement)
    def _(self, statement: MultiStatement):
        for statement in statement.statements:
            self.compile_statement(statement)

    @compile_statement.register(ITE)
    def _(self, ite: ITE):
        if ite.condition.read_global_variables:
            instr = self._create_and_order_instruction(ite)
            self._handle_globals_in_read(instr, ite.condition)
        expr = self.compile_expression(ite.condition)
        self.statement_enablers.append(expr)
        old_parents = self.parent_instructions
        consequence_parents = []
        alternative_parents = []
        self.compile_statement(ite.consequence)
        consequence_parents = self.parent_instructions
        self.parent_instructions = old_parents
        if ite.alternative:
            # flip last condition
            self.statement_enablers[-1] = z3.Not(self.statement_enablers[-1])
            self.compile_statement(ite.alternative)
            alternative_parents = self.parent_instructions
        self.statement_enablers.pop()
        self.parent_instructions = consequence_parents + alternative_parents

    @compile_statement.register(VariableDeclaration)
    def _(self, vardec: VariableDeclaration):
        # WARNING: This only handles local variable declarations
        var = self.create_local_pseuco_variable(vardec.name, vardec.var_type)
        if (vardec.initial_value.read_global_variables):
            instr = self._create_and_order_instruction(vardec)
            self._handle_globals_in_read(instr, vardec.initial_value)
        value = self.compile_expression(vardec.initial_value)
        fact = var == value
        # self.solver.add(fact)
        self.solver.assert_and_track(fact, str(var == value))

    @compile_statement.register(VariableAssignment)
    def _(self, assignment: VariableAssignment):
        assert assignment.is_global_write, "Unsupported modification of local"
        # TODO: cannot handle read and write of global in same statement
        assert not assignment.assignment_expression.read_global_variables, \
            "cannot handle read and write of global in same statement"
        if assignment.assignment_expression.read_global_variables:
            instr = self._create_and_order_instruction(assignment)
            self._handle_globals_in_read(instr,
                                         assignment.assignment_expression)
        written_value = self.compile_expression(
            assignment.assignment_expression)
        instr = self._create_and_order_instruction(assignment)
        self.make_global_write(assignment.var_name, instr, written_value)

    def _track_predecessors(self):
        for instruction in self.instructions:
            predecessor = z3.Const("pre"+str(instruction), Instruction)
            pre, notpre = z3.Consts("pre notpre", Instruction)
            clause = Implies(
                Reachable(instruction),
                Or(
                    z3.Not(Exists([notpre],
                                  InstructionOrder(notpre, instruction))),
                    And(
                        InstructionOrder(pre, instruction),
                        Exists([pre],
                               And(
                                   predecessor == pre,
                                   z3.Not(Exists([notpre],
                                                 And(
                                                     Reachable(notpre),
                                                     notpre != pre,
                                                     InstructionOrder(notpre, instruction),
                                                     InstructionOrder(pre, notpre)
                                                 )
                                                )
                                         )
                               )
                              )
                    )
                )
            )
            self.solver.add(clause)
            self.direct_predecessors[instruction] = predecessor

    def compile(self, program):
        for declaration in program.initialization.variable_declarations:
            self.create_global_pseuco_variable(declaration.name)
            instr = z3.Const(str((self.current_agent_id, declaration.id)),
                             Instruction)
            for existing_instr in self.instructions:
                self.solver.add(instr != existing_instr)
            self.instructions.append(instr)
            # self.solver.add(Reachable(instr))
            self.solver.assert_and_track(Reachable(instr),
                                         "reach %s" % str(instr))
            self.parent_instructions.append(instr)
            expr = self.compile_expression(declaration.initial_value)
            self.make_global_write(declaration.name, instr, expr)
        initial_parents = self.parent_instructions
        for agent_name in itertools.chain.from_iterable(
                program.startedAgentNames.asList()):
            self.parent_instructions = initial_parents
            self.current_agent_id += 1
            self.local_var_table = {}  # reset local variables
            agent = program.initialization.agent_declarations[agent_name]
            body = agent.functionBody
            for statement in body.statements:
                self.compile_statement(statement)
        a = z3.Const('a', Instruction)
        x = z3.Const('x', PseuCoVar)
        p = z3.Const('p', z3.IntSort())
        # specify read and write function
        reads = ForAll([a, x, p],
                       ReadValue(a, x, p) == (
                           Or(*(And(a == b, x == y, p == q)
                                for (b, y, q)
                                in self.global_reads)))
                       )
        writes = ForAll([a, x, p],
                        WriteValue(a, x, p) == (
                            Or(*(And(a == b, x == y, p == q)
                                 for (b, y, q)
                                 in self.global_writes)))
                        )
        # self.solver.add(reads)
        self.solver.assert_and_track(reads, "reads")
        # self.solver.add(writes)
        self.solver.assert_and_track(writes, "writes")
        # self.solver.add(axioms)
        self._track_predecessors()
        sat_stat = self.solver.check()
        assert sat_stat == z3.sat, sat_stat

    def check_sat(self, axioms):
        self.solver.push()
        for axiom in axioms:
            self.solver.assert_and_track(axiom, str(axiom))
        is_sat = self.solver.check()
        if is_sat == z3.unknown:
            self.solver.pop()
            return "Cannot decide given program. WTH did you do?"
        if is_sat == z3.unsat:
            reply = []
            reply.append(100*"_")
            reply.append("UNSAT")
            reply.append(100*"^")
            reply.append(self.solver.unsat_core())
            self.solver.pop()
            return "\n".join(reply)
        counter = 0
        reply = []
        while is_sat == z3.sat and counter < 50:
            counter += 1
            model = self.solver.model()
            reply.append("solution number %d" % counter)
            for (instr, _, value) in self.global_reads:
                reply.append(str(instr) + ": " + str(model[value].as_long()))
            reply.append(100*"-")
            current_model = []
            for (instr, variable, value) in self.global_reads:
                current_model.append(value == model[value].as_long())
            self.solver.assert_and_track(z3.Not(And(*current_model)),
                                         "counter %s" % counter,)
            is_sat = self.solver.check()
        self.solver.pop()
        if counter == 50:
            reply = ["Too many possible reads. Potential thin air read?"]
        return "\n".join(reply)


a, b, c = z3.Consts('a b c', Instruction)
x, y = z3.Consts('x y', PseuCoVar)
p, q = z3.Consts('p q', z3.IntSort())

seqconsist_axioms = [
    ForAll([a, b], Implies(
        And(Reachable(a), Reachable(b)),
        Or(InstructionOrder(a, b), InstructionOrder(b, a))
    )),  # order is total on reachable statements
    ForAll([a, b, c], Implies(
        And(InstructionOrder(a, b), InstructionOrder(b, c)),
        InstructionOrder(a, c)
    )),  # transitive
    ForAll([a, b], Implies(
        And(InstructionOrder(a, b), InstructionOrder(b, a)),
        a == b
    )),
    ForAll([a, x, p], Implies(
        And(ReadValue(a, x, p), Reachable(a)),
        Exists([b], And(WriteValue(b, x, p),
                        Reachable(b),
                        InstructionOrder(b, a),
                        z3.Not(
                            Exists([c, q], And(
                                Reachable(c),
                                c != b,
                                InstructionOrder(b, c),
                                InstructionOrder(c, a),
                                WriteValue(c, x, q)
                            ))
                        )
                        ))
    )),
]

weak_order_axioms = [
    ForAll([a] , Implies(Reachable(a),
                         InstructionOrder(a, a))),  # order is reflexive
    ForAll([a, b, c], Implies(
        And(InstructionOrder(a, b), InstructionOrder(b, c)),
        InstructionOrder(a, c)
    )),  # transitive
    ForAll([a, b], Implies(
        And(InstructionOrder(a, b), InstructionOrder(b, a)),
        a == b
    )),
    ForAll([a, x, p], Implies(
        And(ReadValue(a, x, p), Reachable(a)),
        Exists([b], And(WriteValue(b, x, p),
                        Reachable(b),
                        z3.Not(InstructionOrder(a, b)),
                        z3.Not(
                            Exists([c, q], And(
                                Reachable(c),
                                c != b,
                                InstructionOrder(b, c),
                                InstructionOrder(c, a),
                                WriteValue(c, x, q)
                            ))
                        )
                        ))
    )),
]
